#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
import os
from datetime import datetime
from typing import Dict, List, Optional, Tuple

import breezy.plugins.debian  # For apt: URL support  # noqa: F401
from breezy import urlutils
from breezy.branch import Branch
from breezy.git.repository import GitRepository
from breezy.mutabletree import MutableTree
from breezy.plugins.debian.builder import BuildFailedError
from breezy.plugins.debian.changelog import debcommit
from breezy.plugins.debian.cmds import cmd_builddeb
from breezy.plugins.debian.directory import (
    source_package_vcs,
    vcs_field_to_bzr_url_converters,
)
from breezy.plugins.debian.upstream import MissingUpstreamTarball
from breezy.tree import Tree
from breezy.urlutils import InvalidURL
from breezy.workingtree import WorkingTree
from debian.changelog import Version, get_maintainer
from debian.deb822 import Deb822
from debmutate.changelog import Changelog
from debmutate.changelog import changelog_add_entry as _changelog_add_entry
from debmutate.vcs import split_vcs_url

from .. import workspace as _mod_workspace
from ..probers import select_probers
from ..utils import open_branch

__all__ = [
    "add_changelog_entry",
    "guess_update_changelog",
    "source_package_vcs",
    "build",
    "BuildFailedError",
    "MissingUpstreamTarball",
    "vcs_field_to_bzr_url_converters",
]


DEFAULT_URGENCY = "medium"
DEFAULT_BUILDER = "sbuild --no-clean-source"


class NoSuchPackage(Exception):
    """No such package."""


class NoVcsInformation(Exception):
    """Package does not have any Vcs headers."""


try:
    from lintian_brush.detect_gbp_dch import guess_update_changelog
except ModuleNotFoundError:

    class ChangelogBehaviour:

        def __init__(self, update_changelog, explanation) -> None:
            self.update_changelog = update_changelog
            self.explanation = explanation

        def __str__(self) -> str:
            return self.explanation

        def __repr__(self) -> str:
            return "{}(update_changelog={!r}, explanation={!r})".format(
                type(self).__name__, self.update_changelog,
                self.explanation)

    def guess_update_changelog(
        tree: WorkingTree, debian_path: str, cl: Optional[Changelog] = None
    ) -> Optional[ChangelogBehaviour]:
        logging.warning(
            'Install lintian-brush to detect automatically whether '
            'the changelog should be updated.')
        return ChangelogBehaviour(
            'update_changelog',
            'defaulting to updating changelog since '
            'lintian-brush is not installed')


def add_changelog_entry(
    tree: MutableTree,
    path: str,
    summary: List[str],
    maintainer: Optional[Tuple[str, str]] = None,
    timestamp: Optional[datetime] = None,
    urgency: str = DEFAULT_URGENCY,
) -> None:
    """Add a changelog entry.

    Args:
      tree: Tree to edit
      path: Path to the changelog file
      summary: Entry to add
      maintainer: Maintainer details; tuple of fullname and email
      suppress_warnings: Whether to suppress any warnings from 'dch'
    """
    # TODO(jelmer): This logic should ideally be in python-debian.
    with tree.get_file(path) as f:
        cl = Changelog()
        cl.parse_changelog(
            f, max_blocks=None, allow_empty_author=True, strict=False)
        _changelog_add_entry(
            cl,
            summary=summary,
            maintainer=maintainer,
            timestamp=timestamp,
            urgency=urgency,
        )
    # Workaround until
    # https://salsa.debian.org/python-debian-team/python-debian/-/merge_requests/22
    # lands.
    pieces = []
    for line in cl.initial_blank_lines:
        pieces.append(line.encode(cl._encoding) + b"\n")
    for block in cl._blocks:
        try:
            serialized = block._format(allow_missing_author=True).encode(
                block._encoding
            )
        except TypeError:  # older python-debian
            serialized = bytes(block)
        pieces.append(serialized)
    tree.put_file_bytes_non_atomic(path, b"".join(pieces))


def build(
    tree: WorkingTree,
    subpath: str = "",
    builder: Optional[str] = None,
    result_dir: Optional[str] = None,
) -> None:
    """Build a debian package in a directory.

    Args:
      tree: Working tree
      subpath: Subpath to build in
      builder: Builder command (e.g. 'sbuild', 'debuild')
      result_dir: Directory to copy results to
    """
    if builder is None:
        builder = DEFAULT_BUILDER
    # TODO(jelmer): Refactor brz-debian so it's not necessary
    # to call out to cmd_builddeb, but to lower-level
    # functions instead.
    cmd_builddeb().run(
        [tree.abspath(subpath)], builder=builder, result_dir=result_dir)


def apt_get_source_package(apt_repo, name: str) -> Deb822:
    """Get source package metadata.

    Args:
      name: Name of the source package
    Returns:
      A `Deb822` object
    """
    by_version: Dict[Version, Deb822] = {}

    for source in apt_repo.iter_source_by_name(name):
        by_version[source['Version']] = source

    if len(by_version) == 0:
        raise NoSuchPackage(name)

    # Try the latest version
    version = sorted(by_version, key=Version)[-1]

    return by_version[version]


def convert_debian_vcs_url(vcs_type: str, vcs_url: str) -> str:
    converters = dict(vcs_field_to_bzr_url_converters)
    try:
        return converters[vcs_type](vcs_url)
    except KeyError:
        raise ValueError("unknown vcs %s" % vcs_type)
    except InvalidURL as e:
        raise ValueError("invalid URL: %s" % e)


def open_packaging_branch(
        location, possible_transports=None, vcs_type=None, apt_repo=None):
    """Open a packaging branch from a location string.

    location can either be a package name or a full URL
    """
    if apt_repo is None:
        from breezy.plugins.debian.apt_repo import LocalApt
        apt_repo = LocalApt()
    if "/" not in location and ":" not in location:
        with apt_repo:
            pkg_source = apt_get_source_package(apt_repo, location)
        try:
            (vcs_type, vcs_url) = source_package_vcs(pkg_source)
        except KeyError:
            raise NoVcsInformation(location)
        (url, branch_name, subpath) = split_vcs_url(vcs_url)
    else:
        url, params = urlutils.split_segment_parameters(location)
        try:
            branch_name = urlutils.unquote(params["branch"])
        except KeyError:
            branch_name = None
        subpath = ""
    probers = select_probers(vcs_type)
    branch = open_branch(
        url, possible_transports=possible_transports, probers=probers,
        name=branch_name
    )
    return branch, subpath or ""


def pick_additional_colocated_branches(
        main_branch: Branch) -> Dict[str, str]:
    ret = {
        "pristine-tar": "pristine-tar",
        "pristine-lfs": "pristine-lfs",
        "upstream": "upstream",
        "patch-queue/" + main_branch.name: "patch-queue",  # type: ignore
    }
    if main_branch.name.startswith("debian/"):  # type: ignore
        parts = main_branch.name.split("/")  # type: ignore
        parts[0] = "upstream"
        ret["/".join(parts)] = "upstream"
    existing_branch_names = main_branch.controldir.branch_names()
    return {k: v for (k, v) in ret.items() if k in existing_branch_names}


class Workspace(_mod_workspace.Workspace):
    def __init__(self, main_branch: Branch, *args, **kwargs) -> None:
        if isinstance(main_branch.repository, GitRepository):
            if "additional_colocated_branches" not in kwargs:
                kwargs["additional_colocated_branches"] = {}
            kwargs["additional_colocated_branches"].update(
                pick_additional_colocated_branches(main_branch))
        super().__init__(main_branch, *args, **kwargs)

    @classmethod
    def from_apt_package(cls, package, dir=None):
        main_branch = open_packaging_branch(package)
        return cls(main_branch=main_branch, dir=dir)

    def build(
        self,
        builder: Optional[str] = None,
        result_dir: Optional[str] = None,
        subpath: str = "",
    ) -> None:
        return build(
            tree=self.local_tree,
            subpath=subpath,
            builder=builder,
            result_dir=result_dir,
        )

    def commit(self, message=None, subpath="", paths=None, committer=None,
               reporter=None):
        return debcommit(
            self.local_tree, committer=committer, subpath=subpath,
            paths=paths, reporter=reporter, message=message)


def is_debcargo_package(tree: Tree, subpath: str) -> bool:
    control_path = os.path.join(subpath, "debian", "debcargo.toml")
    return tree.has_filename(control_path)


def control_files_in_root(tree: Tree, subpath: str) -> bool:
    debian_path = os.path.join(subpath, "debian")
    if tree.has_filename(debian_path):
        return False
    control_path = os.path.join(subpath, "control")
    if tree.has_filename(control_path):
        return True
    return tree.has_filename(control_path + ".in")


def _get_maintainer_from_env(env):
    old_env = dict(os.environ.items())
    try:
        os.environ.update(env)
        return get_maintainer()
    finally:
        os.environ.clear()
        os.environ.update(old_env)



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from datetime import datetime

from breezy.tests import TestCase, TestCaseWithTransport
from debian.changelog import ChangelogCreateError

from silver_platter.debian import (
    _get_maintainer_from_env,
    add_changelog_entry,
    convert_debian_vcs_url,
)


class ConvertDebianVcsUrlTests(TestCase):
    def test_git(self):
        self.assertEqual(
            "https://salsa.debian.org/jelmer/blah.git",
            convert_debian_vcs_url(
                "Git", "https://salsa.debian.org/jelmer/blah.git"),
        )

    def test_git_ssh(self):
        self.assertIn(
            convert_debian_vcs_url(
                "Git", "ssh://git@git.kali.org/jelmer/blah.git"),
            ("git+ssh://git@git.kali.org/jelmer/blah.git",
             "ssh://git@git.kali.org/jelmer/blah.git")
        )


class ChangelogAddEntryTests(TestCaseWithTransport):
    def test_edit_existing_new_author(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Initial change.
  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Jane Example")
        self.overrideEnv("DEBEMAIL", "jane@example.com")
        add_changelog_entry(tree, "debian/changelog", ["Add a foo"])
        self.assertFileEqual(
            """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  [ Joe Example ]
  * Initial change.
  * Support updating templated debian/control files that use cdbs
    template.

  [ Jane Example ]
  * Add a foo

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
            "debian/changelog",
        )

    def test_edit_existing_multi_new_author(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  [ Jane Example ]
  * Support updating templated debian/control files that use cdbs
    template.

  [ Joe Example ]
  * Another change

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Jane Example")
        self.overrideEnv("DEBEMAIL", "jane@example.com")
        add_changelog_entry(tree, "debian/changelog", ["Add a foo"])
        self.assertFileEqual(
            """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  [ Jane Example ]
  * Support updating templated debian/control files that use cdbs
    template.

  [ Joe Example ]
  * Another change

  [ Jane Example ]
  * Add a foo

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
            "debian/changelog",
        )

    def test_edit_existing_existing_author(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Joe Example")
        self.overrideEnv("DEBEMAIL", "joe@example.com")
        add_changelog_entry(tree, "debian/changelog", ["Add a foo"])
        self.assertFileEqual(
            """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.
  * Add a foo

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
            "debian/changelog",
        )

    def test_add_new(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) unstable; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Jane Example")
        self.overrideEnv("DEBEMAIL", "jane@example.com")
        self.overrideEnv("DEBCHANGE_VENDOR", "debian")
        add_changelog_entry(
            tree,
            "debian/changelog",
            ["Add a foo"],
            timestamp=datetime(2020, 5, 24, 15, 27, 26),
        )
        self.assertFileEqual(
            """\
lintian-brush (0.36) UNRELEASED; urgency=medium

  * Add a foo

 -- Jane Example <jane@example.com>  Sun, 24 May 2020 15:27:26 -0000

lintian-brush (0.35) unstable; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
            "debian/changelog",
        )

    def test_edit_broken_first_line(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
THIS IS NOT A PARSEABLE LINE
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Jane Example")
        self.overrideEnv("DEBEMAIL", "jane@example.com")
        add_changelog_entry(tree, "debian/changelog", ["Add a foo", "+ Bar"])
        self.assertFileEqual(
            """\
THIS IS NOT A PARSEABLE LINE
lintian-brush (0.35) UNRELEASED; urgency=medium

  [ Joe Example ]
  * Support updating templated debian/control files that use cdbs
    template.

  [ Jane Example ]
  * Add a foo
    + Bar

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
            "debian/changelog",
        )

    def test_add_long_line(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Joe Example")
        self.overrideEnv("DEBEMAIL", "joe@example.com")
        add_changelog_entry(
            tree,
            "debian/changelog",
            [
                "This is adding a very long sentence that is longer than "
                "would fit on a single line in a 80-character-wide line."
            ],
        )
        self.assertFileEqual(
            """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.
  * This is adding a very long sentence that is longer than would fit on a
    single line in a 80-character-wide line.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
            "debian/changelog",
        )

    def test_add_long_subline(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Joe Example")
        self.overrideEnv("DEBEMAIL", "joe@example.com")
        add_changelog_entry(
            tree,
            "debian/changelog",
            [
                "This is the main item.",
                "+ This is adding a very long sentence that is longer than "
                "would fit on a single line in a 80-character-wide line.",
            ],
        )
        self.assertFileEqual(
            """\
lintian-brush (0.35) UNRELEASED; urgency=medium

  * Support updating templated debian/control files that use cdbs
    template.
  * This is the main item.
    + This is adding a very long sentence that is longer than would fit on a
      single line in a 80-character-wide line.

 -- Joe Example <joe@example.com>  Fri, 04 Oct 2019 02:36:13 +0000
""",
            "debian/changelog",
        )

    def test_trailer_only(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) unstable; urgency=medium

  * This line already existed.

 --
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Joe Example")
        self.overrideEnv("DEBEMAIL", "joe@example.com")
        try:
            add_changelog_entry(
                tree, "debian/changelog", ["And this one is new."])
        except ChangelogCreateError:
            self.skipTest(
                "python-debian does not allow serializing changelog "
                "with empty trailer"
            )
        self.assertFileEqual(
            """\
lintian-brush (0.35) unstable; urgency=medium

  * This line already existed.
  * And this one is new.

 --
""",
            "debian/changelog",
        )

    def test_trailer_only_existing_author(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                ("debian/",),
                (
                    "debian/changelog",
                    """\
lintian-brush (0.35) unstable; urgency=medium

  * This line already existed.

  [ Jane Example ]
  * And this one has an existing author.

 --
""",
                ),
            ]
        )
        tree.add(["debian", "debian/changelog"])
        self.overrideEnv("DEBFULLNAME", "Joe Example")
        self.overrideEnv("DEBEMAIL", "joe@example.com")
        try:
            add_changelog_entry(
                tree, "debian/changelog", ["And this one is new."])
        except ChangelogCreateError:
            self.skipTest(
                "python-debian does not allow serializing changelog "
                "with empty trailer"
            )
        self.assertFileEqual(
            """\
lintian-brush (0.35) unstable; urgency=medium

  * This line already existed.

  [ Jane Example ]
  * And this one has an existing author.

  [ Joe Example ]
  * And this one is new.

 --
""",
            "debian/changelog",
        )


class GetMaintainerFromEnvTests(TestCase):

    def test_normal(self):
        t = _get_maintainer_from_env({})
        self.assertIsInstance(t, tuple)
        self.assertIsInstance(t[0], str)
        self.assertIsInstance(t[1], str)

    def test_env(self):
        t = _get_maintainer_from_env({
            'DEBFULLNAME': 'Jelmer',
            'DEBEMAIL': 'jelmer@example.com',
        })
        self.assertEqual(('Jelmer', 'jelmer@example.com'), t)



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import subprocess
from typing import Dict, Optional

from breezy.branch import Branch
from breezy.revision import RevisionID
from breezy.workingtree import WorkingTree
from . import _svp_rs

create_temp_sprout = _svp_rs.create_temp_sprout


class TemporarySprout:
    """Create a temporary sprout of a branch.

    This attempts to fetch the least amount of history as possible.
    """

    def __init__(
        self,
        branch: Branch,
        additional_colocated_branches: Optional[Dict[str, str]] = None,
        dir: Optional[str] = None,
    ) -> None:
        self.branch = branch
        self.additional_colocated_branches = additional_colocated_branches
        self.dir = dir

    def __enter__(self) -> WorkingTree:
        tree, self._destroy = create_temp_sprout(
            self.branch,
            additional_colocated_branches=self.additional_colocated_branches,
            dir=self.dir,
        )
        return tree

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._destroy()
        return False


class PreCheckFailed(Exception):
    """The post check failed."""


def run_pre_check(tree: WorkingTree, script: Optional[str]) -> None:
    """Run a script ahead of making any changes to a tree.

    Args:
      tree: The working tree to operate in
      script: Command to run
    Raises:
      PreCheckFailed: If the pre-check failed
    """
    if not script:
        return
    try:
        subprocess.check_call(script, shell=True, cwd=tree.basedir)
    except subprocess.CalledProcessError:
        raise PreCheckFailed()


class PostCheckFailed(Exception):
    """The post check failed."""


def run_post_check(
    tree: WorkingTree, script: Optional[str], since_revid: RevisionID
) -> None:
    """Run a script after making any changes to a tree.

    Args:
      tree: The working tree to operate in
      script: Command to run
      since_revid: Revision id since which changes were made
    Raises:
      PostCheckFailed: If the pre-check failed
    """
    if not script:
        return
    try:
        subprocess.check_call(
            script, shell=True, cwd=tree.basedir,
            env={"SINCE_REVID": since_revid}
        )
    except subprocess.CalledProcessError:
        raise PostCheckFailed()


BranchTemporarilyUnavailable = _svp_rs.BranchTemporarilyUnavailable
BranchRateLimited = _svp_rs.BranchRateLimited
BranchUnavailable = _svp_rs.BranchUnavailable
BranchMissing = _svp_rs.BranchMissing
BranchUnsupported = _svp_rs.BranchUnsupported
open_branch = _svp_rs.open_branch
open_branch_containing = _svp_rs.open_branch_containing
full_branch_url = _svp_rs.full_branch_url


def get_branch_vcs_type(branch):
    vcs = getattr(branch.repository, "vcs", None)
    if vcs:
        return vcs.abbreviation
    else:
        return "bzr"



#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from contextlib import suppress
from typing import List, Optional, Type

from breezy.bzr import RemoteBzrProber
from breezy.controldir import ControlDirFormat, Prober
from breezy.errors import UnsupportedFormatError
from breezy.git import RemoteGitProber


class UnsupportedVCSProber(Prober):
    def __init__(self, vcs_type) -> None:
        self.vcs_type = vcs_type

    def __eq__(self, other):
        return (isinstance(other, type(self))
                and other.vcs_type == self.vcs_type)

    def __call__(self):
        # The prober expects to be registered as a class.
        return self

    def priority(self, transport):
        return 200

    def probe_transport(self, transport):
        raise UnsupportedFormatError(
            "This VCS %s is not currently supported." % self.vcs_type
        )

    @classmethod
    def known_formats(klass):
        return []


prober_registry = {
    "bzr": RemoteBzrProber,
    "git": RemoteGitProber,
}

try:
    from breezy.plugins.fossil import RemoteFossilProber
except ImportError:
    pass
else:
    prober_registry["fossil"] = RemoteFossilProber

try:
    from breezy.plugins.svn import SvnRepositoryProber
except ImportError:
    pass
else:
    prober_registry["svn"] = SvnRepositoryProber

try:
    from breezy.plugins.hg import SmartHgProber
except ImportError:
    pass
else:
    prober_registry["hg"] = SmartHgProber

try:
    from breezy.plugins.darcs import DarcsProber
except ImportError:
    pass
else:
    prober_registry["darcs"] = DarcsProber

try:
    from breezy.plugins.cvs import CVSProber
except ImportError:
    pass
else:
    prober_registry["cvs"] = CVSProber


def select_probers(vcs_type=None):
    if vcs_type is None:
        return None
    try:
        return [prober_registry[vcs_type.lower()]]
    except KeyError:
        return [UnsupportedVCSProber(vcs_type)]


def select_preferred_probers(
        vcs_type: Optional[str] = None) -> List[Type[Prober]]:
    probers = list(ControlDirFormat.all_probers())
    if vcs_type:
        with suppress(KeyError):
            probers.insert(0, prober_registry[vcs_type.lower()])
    return probers



#!/usr/bin/python
# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from typing import List, Optional

from breezy import errors
from breezy import merge as _mod_merge  # type: ignore
from breezy import revision as _mod_revision
from breezy.branch import Branch
from breezy.errors import PermissionDenied
from breezy.forge import (
    NoSuchProject,
    SourceNotDerivedFromTarget,
    UnsupportedForge,
    MergeProposal,
    MergeProposalExists,
)
from breezy.memorybranch import MemoryBranch
from breezy.revision import RevisionID

from . import _svp_rs

__all__ = [
    "push_changes",
    "push_derived_changes",
    "propose_changes",
    "EmptyMergeProposal",
    "check_proposal_diff",
    "find_existing_proposed",
    "NoSuchProject",
    "PermissionDenied",
    "UnsupportedForge",
    "SourceNotDerivedFromTarget",
    "MergeProposal",
    "MergeProposalExists",
]


MODE_PUSH = "push"
MODE_ATTEMPT_PUSH = "attempt-push"
MODE_PROPOSE = "propose"
MODE_PUSH_DERIVED = "push-derived"
SUPPORTED_MODES: List[str] = [
    MODE_PUSH,
    MODE_ATTEMPT_PUSH,
    MODE_PROPOSE,
    MODE_PUSH_DERIVED,
]


push_result = _svp_rs.push_result
push_changes = _svp_rs.push_changes
push_derived_changes = _svp_rs.push_derived_changes
propose_changes = _svp_rs.propose_changes
publish_changes = _svp_rs.publish_changes
PublishResult = _svp_rs.PublishResult
InsufficientChangesForNewProposal = _svp_rs.InsufficientChangesForNewProposal


class EmptyMergeProposal(Exception):
    """Merge proposal does not have any changes."""

    def __init__(self, local_branch: Branch, main_branch: Branch) -> None:
        self.local_branch = local_branch
        self.main_branch = main_branch


def check_proposal_diff(
    other_branch: Branch, main_branch: Branch,
    stop_revision: Optional[RevisionID] = None
) -> None:
    if stop_revision is None:
        stop_revision = other_branch.last_revision()
    main_revid = main_branch.last_revision()
    other_branch.repository.fetch(main_branch.repository, main_revid)
    with other_branch.lock_read():
        main_tree = other_branch.repository.revision_tree(main_revid)
        revision_graph = other_branch.repository.get_graph()
        tree_branch = MemoryBranch(
            other_branch.repository, (None, main_revid), None)
        merger = _mod_merge.Merger(
            tree_branch, this_tree=main_tree, revision_graph=revision_graph
        )
        merger.set_other_revision(stop_revision, other_branch)
        try:
            merger.find_base()
        except errors.UnrelatedBranches:
            merger.set_base_revision(_mod_revision.NULL_REVISION, other_branch)
        merger.merge_type = _mod_merge.Merge3Merger  # type: ignore
        tree_merger = merger.make_merger()
        with tree_merger.make_preview_transform() as tt:
            changes = tt.iter_changes()
            if not any(changes):
                raise EmptyMergeProposal(other_branch, main_branch)


find_existing_proposed = _svp_rs.find_existing_proposed


def merge_conflicts(
    main_branch: Branch, other_branch: Branch,
    other_revision: Optional[RevisionID] = None
) -> bool:
    """Check whether two branches are conflicted when merged.

    Args:
      main_branch: Main branch to merge into
      other_branch: Branch to merge (and use for scratch access, needs write
                    access)
      other_revision: Other revision to check
    Returns:
      boolean indicating whether the merge would result in conflicts
    """
    if other_revision is None:
        other_revision = other_branch.last_revision()
    if other_branch.repository.get_graph().is_ancestor(
        main_branch.last_revision(), other_revision
    ):
        return False

    other_branch.repository.fetch(
        main_branch.repository, revision_id=main_branch.last_revision()
    )

    # Reset custom merge hooks, since they could make it harder to detect
    # conflicted merges that would appear on the hosting site.
    old_file_content_mergers = _mod_merge.Merger.hooks["merge_file_content"]
    _mod_merge.Merger.hooks["merge_file_content"] = []

    other_tree = other_branch.repository.revision_tree(other_revision)
    try:
        try:
            merger = _mod_merge.Merger.from_revision_ids(
                other_tree,
                other_branch=other_branch,
                other=main_branch.last_revision(),
                tree_branch=other_branch,
            )
        except errors.UnrelatedBranches:
            # Unrelated branches don't technically *have* to lead to
            # conflicts, but there's not a lot to be salvaged here, either.
            return True
        merger.merge_type = _mod_merge.Merge3Merger
        tree_merger = merger.make_merger()
        with tree_merger.make_preview_transform():
            return bool(tree_merger.cooked_conflicts)
    finally:
        _mod_merge.Merger.hooks["merge_file_content"] = (
            old_file_content_mergers)



#!/usr/bin/python
# Copyright (C) 2021 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import json
import logging
import os
import subprocess
import sys
import tempfile
from contextlib import suppress
from dataclasses import dataclass, field
from typing import Any, BinaryIO, Dict, List, Optional, Tuple, Union

from breezy.commit import PointlessCommit
from breezy.revision import RevisionID
from breezy.workingtree import WorkingTree
from breezy.workspace import check_clean_tree, reset_tree
from debian.changelog import Changelog
from debian.deb822 import Deb822

from ..apply import (
    ResultFileFormatError,
    ScriptFailed,
    ScriptMadeNoChanges,
    ScriptNotFound,
)
from . import (
    DEFAULT_BUILDER,
    BuildFailedError,
    MissingUpstreamTarball,
    _get_maintainer_from_env,
    add_changelog_entry,
    build,
    control_files_in_root,
    guess_update_changelog,
)


class MissingChangelog(Exception):
    """No changelog file is present."""


class DetailedFailure(Exception):
    """Detailed failure."""

    def __init__(self, source_name, result_code, description, stage=None,
                 details=None) -> None:
        self.source = source_name
        self.result_code = result_code
        self.description = description
        self.details = details
        self.stage = stage

    @classmethod
    def from_json(cls, source_name, json):
        return cls(
            source_name,
            result_code=json.get('result_code'),
            description=json.get('description'),
            stage=tuple(json['stage']) if json.get('stage') else None,
            details=json.get('details'))


@dataclass
class CommandResult:

    source: Optional[str]
    description: Optional[str] = None
    value: Optional[int] = None
    serialized_context: Optional[str] = None
    context: Dict[str, str] = field(default_factory=dict)
    tags: List[Tuple[str, RevisionID]] = field(default_factory=list)
    old_revision: Optional[RevisionID] = None
    new_revision: Optional[RevisionID] = None
    target_branch_url: Optional[str] = None

    @classmethod
    def from_json(cls, source, data):
        if 'tags' in data:
            tags = []
            for name, revid in data['tags']:
                tags.append((name, revid.encode('utf-8')))
        else:
            tags = None
        return cls(
            source=source,
            value=data.get('value', None),
            context=data.get('context', {}),
            serialized_context=data.get('serialized_context', None),
            description=data.get('description'),
            target_branch_url=data.get('target-branch-url', None),
            tags=tags)


def install_built_package(local_tree, subpath, build_target_dir):
    import re
    import subprocess
    abspath = local_tree.abspath(os.path.join(subpath, 'debian/changelog'))
    with open(abspath) as f:
        cl = Changelog(f)
    non_epoch_version = cl[0].version.upstream_version
    if cl[0].version.debian_version is not None:
        non_epoch_version += "-%s" % cl[0].version.debian_version
    c = re.compile(
        '{}_{}_(.*).changes'.format(
            re.escape(cl[0].package),
            re.escape(non_epoch_version)))  # type: ignore
    for entry in os.scandir(build_target_dir):
        if not c.match(entry.name):
            continue
        with open(entry.path, 'rb') as g:
            changes = Deb822(g)
            if changes.get('Binary'):
                subprocess.check_call(['debi', entry.path])


def script_runner(   # noqa: C901
    local_tree: WorkingTree, script: Union[str, List[str]],
    commit_pending: Optional[bool] = None,
    resume_metadata: Optional[Any] = None,
    subpath: str = '', update_changelog: Optional[bool] = None,
    extra_env: Optional[Dict[str, str]] = None,
    committer: Optional[str] = None,
    stderr: Optional[BinaryIO] = None
) -> CommandResult:  # noqa: C901
    """Run a script in a tree and commit the result.

    This ignores newly added files.

    Args:
      local_tree: Local tree to run script in
      script: Script to run
      commit_pending: Whether to commit pending changes
        (True, False or None: only commit if there were no commits by the
         script)
    """
    if control_files_in_root(local_tree, subpath):
        debian_path = subpath
    else:
        debian_path = os.path.join(subpath, "debian")
    if update_changelog is None:
        dch_guess = guess_update_changelog(local_tree, debian_path)
        if dch_guess:
            if isinstance(dch_guess, tuple):  # lintian-brush < 1.22
                update_changelog, explanation = dch_guess
            else:
                update_changelog = dch_guess.update_changelog
                explanation = dch_guess.explanation
            logging.info('%s', explanation)
        else:
            # Assume yes.
            update_changelog = True

    cl_path = os.path.join(debian_path, 'changelog')
    try:
        with open(local_tree.abspath(cl_path)) as f:
            cl = Changelog(f)
            source_name = cl[0].package
    except FileNotFoundError:
        source_name = None

    env = dict(os.environ)
    if extra_env:
        env.update(extra_env)

    env['SVP_API'] = '1'
    if source_name:
        env['DEB_SOURCE'] = source_name

    if update_changelog:
        env['DEB_UPDATE_CHANGELOG'] = 'update'
    else:
        env['DEB_UPDATE_CHANGELOG'] = 'leave'

    last_revision = local_tree.last_revision()
    orig_tags = local_tree.branch.tags.get_tag_dict()
    with tempfile.TemporaryDirectory() as td:
        env['SVP_RESULT'] = os.path.join(td, 'result.json')
        if resume_metadata:
            env['SVP_RESUME'] = os.path.join(td, 'resume-metadata.json')
            with open(env['SVP_RESUME'], 'w') as f:
                json.dump(resume_metadata, f)
        try:
            p = subprocess.Popen(
                script, cwd=local_tree.abspath(subpath),
                stdout=subprocess.PIPE,
                shell=isinstance(script, str), env=env,
                stderr=stderr)
        except FileNotFoundError as e:
            raise ScriptNotFound(script) from e
        (description_encoded, err) = p.communicate(b"")
        try:
            with open(env['SVP_RESULT']) as f:
                try:
                    result_json = json.load(f)
                except json.decoder.JSONDecodeError as e:
                    raise ResultFileFormatError(e)
        except FileNotFoundError:
            result_json = None
        if p.returncode != 0:
            if result_json is not None:
                raise DetailedFailure.from_json(source_name, result_json)
            raise ScriptFailed(script, p.returncode)
        # If the changelog didn't exist earlier, then hopefully it was created
        # now.
        if source_name is None:
            try:
                with open(local_tree.abspath(cl_path)) as f:
                    cl = Changelog(f)
                    source_name = cl[0].package
            except FileNotFoundError:
                raise MissingChangelog(cl_path)
        if result_json is not None:
            result = CommandResult.from_json(source_name, result_json)
        else:
            result = CommandResult(source=source_name)
    if not result.description:
        result.description = description_encoded.decode().replace("\r", "")
    new_revision = local_tree.last_revision()
    if result.tags is None:
        result.tags = []
        for name, revid in local_tree.branch.tags.get_tag_dict().items():
            if orig_tags.get(name) != revid:
                result.tags.append((name, revid))
    if last_revision == new_revision and commit_pending is None:
        # Automatically commit pending changes if the script did not
        # touch the branch.
        commit_pending = True
    if commit_pending:
        if (update_changelog
                and result.description
                and local_tree.has_changes()):
            add_changelog_entry(
                local_tree,
                os.path.join(debian_path, 'changelog'),
                [result.description],
                maintainer=_get_maintainer_from_env(extra_env))
        local_tree.smart_add([local_tree.abspath(subpath)])
        with suppress(PointlessCommit):
            new_revision = local_tree.commit(
                result.description, allow_pointless=False,
                committer=committer)
    if new_revision == last_revision:
        raise ScriptMadeNoChanges()
    result.old_revision = last_revision
    result.new_revision = new_revision
    return result


def main(argv: List[str]) -> Optional[int]:  # noqa: C901
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--command", help="Path to script to run.", type=str)
    parser.add_argument(
        "--diff", action="store_true", help="Show diff of generated changes."
    )
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=None,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )

    parser.add_argument(
        "--commit-pending",
        help="Commit pending changes after script.",
        choices=["yes", "no", "auto"],
        default=None,
        type=str,
    )
    parser.add_argument(
        "--build-verify",
        help="Build package to verify it.",
        dest="build_verify",
        action="store_true",
    )
    parser.add_argument(
        "--builder",
        default=DEFAULT_BUILDER,
        type=str,
        help="Build command to use when verifying build.",
    )
    parser.add_argument(
        "--build-target-dir",
        type=str,
        help=(
            "Store built Debian files in specified directory "
            "(with --build-verify)"
        ),
    )
    parser.add_argument(
        "--install", "-i",
        action="store_true",
        help="Install built package (implies --build-verify)")
    parser.add_argument(
        "--dump-context", action="store_true",
        help="Report context on success")

    parser.add_argument(
        "--recipe", type=str, help="Recipe to use.")
    args = parser.parse_args(argv)

    if args.recipe:
        from ..recipe import Recipe
        recipe = Recipe.from_path(args.recipe)
    else:
        recipe = None

    if args.commit_pending:
        commit_pending = {
            "auto": None, "yes": True, "no": False}[args.commit_pending]
    elif recipe:
        commit_pending = recipe.commit_pending
    else:
        commit_pending = None

    if args.command:
        command = args.command
    elif recipe and recipe.command:
        command = recipe.command
    else:
        parser.error('No command or recipe specified.')

    local_tree, subpath = WorkingTree.open_containing('.')

    check_clean_tree(local_tree)

    try:
        try:
            result = script_runner(
                local_tree, script=command, commit_pending=commit_pending,
                subpath=subpath, update_changelog=args.update_changelog)
        except MissingChangelog as e:
            logging.error('No debian changelog file (%s) present', e.args[0])
            return False
        except ScriptMadeNoChanges:
            logging.info('Script made no changes')
            return False

        if result.description:
            logging.info('Succeeded: %s', result.description)

        if args.build_verify or args.install:
            try:
                build(local_tree, subpath, builder=args.builder,
                      result_dir=args.build_target_dir)
            except BuildFailedError:
                logging.error("%s: build failed", result.source)
                return False
            except MissingUpstreamTarball:
                logging.error(
                    "%s: unable to find upstream source", result.source)
                return False
    except Exception:
        reset_tree(local_tree, subpath=subpath)
        raise

    if args.install:
        install_built_package(local_tree, subpath, args.build_target_dir)

    if args.diff:
        from breezy.diff import show_diff_trees
        old_tree = local_tree.revision_tree(result.old_revision)
        new_tree = local_tree.revision_tree(result.new_revision)
        show_diff_trees(
            old_tree,
            new_tree,
            sys.stdout.buffer,
            old_label='old/',
            new_label='new/')

    if args.dump_context:
        json.dump(result.context, sys.stdout, indent=5)
    return 0



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy.tests import TestCaseWithTransport

from silver_platter.publish import (
    EmptyMergeProposal,
    check_proposal_diff,
    push_result,
)


class PushResultTests(TestCaseWithTransport):
    def test_simple(self):
        target = self.make_branch("target")
        source = self.make_branch_and_tree("source")
        revid = source.commit("Some change")
        push_result(source.branch, target)
        self.assertEqual(target.last_revision(), revid)


class CheckProposalDiffBase:
    def test_no_new_commits(self):
        orig = self.make_branch_and_tree("orig", format=self.format)
        self.build_tree(["orig/a"])
        orig.add(["a"])
        orig.commit("blah")

        proposal = orig.controldir.sprout("proposal").open_branch()

        self.addCleanup(proposal.lock_write().unlock)
        self.assertRaises(
            EmptyMergeProposal, check_proposal_diff, proposal, orig.branch
        )

    def test_no_op_commits(self):
        orig = self.make_branch_and_tree("orig", format=self.format)
        self.build_tree(["orig/a"])
        orig.add(["a"])
        orig.commit("blah")

        proposal = orig.controldir.sprout("proposal").open_workingtree()
        proposal.commit("another commit that is pointless")

        self.addCleanup(proposal.lock_write().unlock)
        self.assertRaises(
            EmptyMergeProposal, check_proposal_diff, proposal.branch,
            orig.branch
        )

    def test_indep(self):
        orig = self.make_branch_and_tree("orig", format=self.format)
        self.build_tree(["orig/a"])
        orig.add(["a"])
        orig.commit("blah")

        proposal = orig.controldir.sprout("proposal").open_workingtree()
        self.build_tree_contents([("orig/b", "b"), ("orig/c", "c")])
        orig.add(["b", "c"])
        orig.commit("independent")

        self.build_tree_contents([("proposal/b", "b")])
        if proposal.supports_setting_file_ids():
            proposal.add(["b"], ids=[orig.path2id("b")])
        else:
            proposal.add(["b"])
        proposal.commit("not pointless")

        self.addCleanup(proposal.lock_write().unlock)
        self.assertRaises(
            EmptyMergeProposal, check_proposal_diff, proposal.branch,
            orig.branch)

    def test_changes(self):
        orig = self.make_branch_and_tree("orig", format=self.format)
        self.build_tree(["orig/a"])
        orig.add(["a"])
        orig.commit("blah")

        proposal = orig.controldir.sprout("proposal").open_workingtree()
        self.build_tree(["proposal/b"])
        proposal.add(["b"])
        proposal.commit("not pointless")

        self.addCleanup(proposal.lock_write().unlock)
        check_proposal_diff(proposal.branch, orig.branch)


class CheckProposalDiffGitTests(TestCaseWithTransport, CheckProposalDiffBase):

    format = "git"


class CheckProposalDiffBzrTests(TestCaseWithTransport, CheckProposalDiffBase):

    format = "bzr"



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for uploading packages."""

import datetime
import logging
import os
import subprocess
import sys
import tempfile
import time
from contextlib import suppress
from email.utils import parseaddr
from typing import Callable, List, Optional, Tuple

from breezy import gpg  # type: ignore
from breezy.commit import NullCommitReporter, PointlessCommit
from breezy.config import NoEmailInUsername, extract_email_address
from breezy.errors import NoSuchTag, PermissionDenied
from breezy.plugins.debian.apt_repo import Apt, LocalApt, RemoteApt
from breezy.plugins.debian.builder import BuildFailedError
from breezy.plugins.debian.cmds import _build_helper
from breezy.plugins.debian.import_dsc import DistributionBranch
from breezy.plugins.debian.release import release
from breezy.plugins.debian.upstream import MissingUpstreamTarball
from breezy.plugins.debian.util import (
    MissingChangelogError,
    NoPreviousUpload,
    changelog_find_previous_upload,
    dput_changes,
    find_changelog,
)
from breezy.revision import NULL_REVISION
from breezy.tree import MissingNestedTree
from breezy.workingtree import WorkingTree
from debian.changelog import Version, get_maintainer
from debmutate.changelog import (
    ChangelogEditor,
    ChangelogParseError,
    changeblock_ensure_first_line,
    gbp_dch,
)
from debmutate.control import ControlEditor
from debmutate.reformatting import GeneratedFile

import silver_platter  # noqa: F401

from ..probers import select_probers
from ..utils import (
    BranchMissing,
    BranchRateLimited,
    BranchUnavailable,
    BranchUnsupported,
    open_branch,
)
from . import (
    DEFAULT_BUILDER,
    NoSuchPackage,
    Workspace,
    apt_get_source_package,
    source_package_vcs,
    split_vcs_url,
)


def debsign(path, keyid=None):
    (bd, changes_file) = os.path.split(path)
    args = ["debsign"]
    if keyid:
        args.append("-k%s" % keyid)
    args.append(changes_file)
    subprocess.check_call(args, cwd=bd)


class LastUploadMoreRecent(Exception):
    """Last version in archive is newer than vcs version."""

    def __init__(self, archive_version, vcs_version) -> None:
        self.archive_version = archive_version
        self.vcs_version = vcs_version
        super().__init__(
            f"last upload ({archive_version}) is more recent "
            f"than vcs ({vcs_version})"
        )


class NoUnuploadedChanges(Exception):
    """Indicates there are no unuploaded changes for a package."""

    def __init__(self, archive_version) -> None:
        self.archive_version = archive_version
        super().__init__(
            "nothing to upload, latest version is in archive: %s" %
            archive_version
        )


class NoUnreleasedChanges(Exception):
    """Indicates there are no unreleased changes for a package."""

    def __init__(self, version) -> None:
        self.version = version
        super().__init__(
            "nothing to upload, latest version in vcs is not unreleased: %s" %
            version
        )


class RecentCommits(Exception):
    """Indicates there are too recent commits for a package."""

    def __init__(self, commit_age, min_commit_age) -> None:
        self.commit_age = commit_age
        self.min_commit_age = min_commit_age
        super().__init__(
            "Last commit is only %d days old (< %d)"
            % (self.commit_age, self.min_commit_age)
        )


class CommitterNotAllowed(Exception):
    """Specified committer is not allowed."""

    def __init__(self, committer, allowed_committers) -> None:
        self.committer = committer
        self.allowed_committers = allowed_committers
        super().__init__(
            f"Committer {self.committer} not in allowed committers: "
            f"{self.allowed_committers!r}"
        )


class LastReleaseRevisionNotFound(Exception):
    """The revision for the last uploaded release can't be found."""

    def __init__(self, package, version) -> None:
        self.package = package
        self.version = version
        super().__init__(
            f"Unable to find revision matching version {version!r} "
            f"for {package}"
        )


def check_revision(rev, min_commit_age, allowed_committers):
    """Check whether a revision can be included in an upload.

    Args:
      rev: revision to check
      min_commit_age: Minimum age for revisions
      allowed_committers: List of allowed committers
    Raises:
      RecentCommits: When there are commits younger than min_commit_age
    """
    # TODO(jelmer): deal with timezone
    if min_commit_age is not None:
        commit_time = datetime.datetime.fromtimestamp(rev.timestamp)
        time_delta = datetime.datetime.now() - commit_time
        if time_delta.days < min_commit_age:
            raise RecentCommits(time_delta.days, min_commit_age)
    # TODO(jelmer): Allow tag to prevent automatic uploads
    try:
        committer_email = extract_email_address(rev.committer)
    except NoEmailInUsername as e:
        logging.warning(
            'Unable to extract email from %r', rev.committer)
        raise CommitterNotAllowed(rev.committer, allowed_committers) from e
    if allowed_committers and committer_email not in allowed_committers:
        raise CommitterNotAllowed(committer_email, allowed_committers)


def find_last_release_revid(branch, version):
    db = DistributionBranch(branch, None)
    return db.revid_of_version(version)


def get_maintainer_keys(context):
    for key in context.keylist(
            source="/usr/share/keyrings/debian-keyring.gpg"):
        yield key.fpr
        for subkey in key.subkeys:
            yield subkey.keyid


class GbpDchFailed(Exception):
    """gbp dch failed to run."""


class GeneratedChangelogFile(Exception):
    """unable to update changelog since it is generated."""


def prepare_upload_package(  # noqa: C901
    local_tree: WorkingTree,
    subpath: str,
    pkg: str,
    last_uploaded_version: Optional[Version],
    builder: str,
    *,
    gpg_strategy: Optional[gpg.GPGStrategy] = None,
    min_commit_age: Optional[int] = None,
    allowed_committers: Optional[List[str]] = None,
    apt: Optional[Apt] = None,
) -> Tuple[str, str]:
    debian_path = os.path.join(subpath, "debian")
    try:
        from lintian_brush.detect_gbp_dch import guess_update_changelog
    except ImportError:
        run_gbp_dch = True   # Let's just try
    else:
        cl_behaviour = guess_update_changelog(local_tree, debian_path)
        run_gbp_dch = (
            cl_behaviour is None or not cl_behaviour.update_changelog)
    if run_gbp_dch:
        try:
            gbp_dch(local_tree.abspath(subpath))
        except subprocess.CalledProcessError:
            # TODO(jelmer): gbp dch sometimes fails when there is no existing
            # open changelog entry; it fails invoking
            # "dpkg --lt None <old-version>"
            raise GbpDchFailed()
        local_tree.commit(
            specific_files=[os.path.join(debian_path, 'changelog')],
            message='update changelog\n\nGbp-Dch: Ignore')
    cl, top_level = find_changelog(local_tree, merge=False, max_blocks=None)
    if (last_uploaded_version is not None
            and cl.version == last_uploaded_version):
        raise NoUnuploadedChanges(cl.version)
    try:
        previous_version_in_branch = changelog_find_previous_upload(cl)
    except NoPreviousUpload:
        pass
    else:
        if (last_uploaded_version is not None and
                last_uploaded_version > previous_version_in_branch):
            raise LastUploadMoreRecent(
                last_uploaded_version, previous_version_in_branch)

    if last_uploaded_version is not None:
        logging.info("Checking revisions since %s", last_uploaded_version)
    with local_tree.lock_read():
        if last_uploaded_version is not None:
            try:
                last_release_revid = find_last_release_revid(
                    local_tree.branch, last_uploaded_version
                )
            except NoSuchTag:
                raise LastReleaseRevisionNotFound(pkg, last_uploaded_version)
        else:
            last_release_revid = NULL_REVISION
        graph = local_tree.branch.repository.get_graph()
        revids = list(
            graph.iter_lefthand_ancestry(
                local_tree.branch.last_revision(), [last_release_revid]
            )
        )
        if not revids:
            logging.info("No pending changes")
            raise NoUnuploadedChanges(cl.version)
        if gpg_strategy:
            logging.info("Verifying GPG signatures...")
            count, result, all_verifiables = gpg.bulk_verify_signatures(
                local_tree.branch.repository, revids, gpg_strategy
            )
            for revid, code, _key in result:
                if code != gpg.SIGNATURE_VALID:
                    raise Exception(
                        "No valid GPG signature on %r: %d" % (revid, code))
        for _revid, rev in local_tree.branch.repository.iter_revisions(revids):
            if rev is not None:
                check_revision(rev, min_commit_age, allowed_committers)

        if cl.distributions != "UNRELEASED":
            raise NoUnreleasedChanges(cl.version)
    qa_upload = False
    team_upload = False
    control_path = local_tree.abspath(os.path.join(debian_path, "control"))
    with ControlEditor(control_path) as e:
        maintainer = parseaddr(e.source["Maintainer"])
        if maintainer[1] == "packages@qa.debian.org":
            qa_upload = True
        # TODO(jelmer): Check whether this is a team upload
        # TODO(jelmer): determine whether this is a NMU upload
    if qa_upload or team_upload:
        changelog_path = local_tree.abspath(
            os.path.join(debian_path, "changelog"))
        with ChangelogEditor(changelog_path) as e:
            if qa_upload:
                changeblock_ensure_first_line(e[0], "QA upload.")
                message = "Mention QA Upload."
            elif team_upload:
                changeblock_ensure_first_line(e[0], "Team upload.")
                message = "Mention Team Upload."
            else:
                message = None
        if message is not None:
            with suppress(PointlessCommit):
                local_tree.commit(
                    specific_files=[os.path.join(debian_path, "changelog")],
                    message=message,
                    allow_pointless=False,
                    reporter=NullCommitReporter(),
                )
    try:
        tag_name = release(local_tree, subpath)
    except GeneratedFile:
        raise GeneratedChangelogFile()
    target_dir = tempfile.mkdtemp()
    if last_uploaded_version is not None:
        builder = builder.replace(
            "${LAST_VERSION}", str(last_uploaded_version))
    target_changes = _build_helper(
        local_tree, subpath, local_tree.branch, target_dir, builder=builder,
        apt=apt
    )
    debsign(target_changes['source'])
    return target_changes['source'], tag_name


def select_apt_packages(apt_repo, package_names, maintainer):
    packages = []

    with apt_repo:
        for source in apt_repo.iter_sources():
            if maintainer:
                fullname, email = parseaddr(source['Maintainer'])
                if email not in maintainer:
                    continue

            if package_names and source['Package'] not in package_names:
                continue

            packages.append(source['Package'])

        return packages


class PackageProcessingFailure(Exception):

    def __init__(self, reason, description=None) -> None:
        self.reason = reason
        self.description = description


class PackageIgnored(Exception):

    def __init__(self, reason, description=None) -> None:
        self.reason = reason
        self.description = description


def check_git_commits(vcslog, min_commit_age, allowed_committers):
    class GitRevision:

        @property
        def committer(self):
            return self.headers.get('Committer') or self.headers.get('Author')

        @property
        def timestamp(self):
            datestr = self.headers.get('Date')
            dt = datetime.datetime.strptime(
                datestr.strip(), "%a %b %d %H:%M:%S %Y %z")
            return dt.timestamp()

        def __init__(self, commit_id, headers, message) -> None:
            self.commit_id = commit_id
            self.headers = headers
            self.message = message

        @classmethod
        def from_lines(cls, lines):
            commit_id = None
            message = []
            headers = {}
            for i, line in enumerate(lines):
                if line.startswith('commit '):
                    commit_id = line[len('commit '):]
                elif line == '':
                    message = lines[i+1:]
                    break
                else:
                    name, value = line.split(': ')
                    headers[name] = value
            return cls(commit_id, headers, message)

    last_commit_ts = None
    lines = []
    for line in vcslog.splitlines():
        if line == '' and lines[-1][0].isspace():
            gitrev = GitRevision.from_lines(lines)
            if last_commit_ts is None:
                last_commit_ts = gitrev.timestamp
            check_revision(
                gitrev, min_commit_age,
                allowed_committers)
            lines = []
        else:
            lines.append(line)
    if lines:
        gitrev = GitRevision.from_lines(lines)
        if last_commit_ts is None:
            last_commit_ts = gitrev.timestamp
        check_revision(gitrev, min_commit_age, allowed_committers)
    return last_commit_ts


def process_package(
        apt_repo, package,
        builder: str, *, exclude=None, autopkgtest_only: bool = False,
        gpg_verification: bool = False,
        acceptable_keys=None, debug: bool = False,
        diff: bool = False, min_commit_age=None, allowed_committers=None,
        vcs_type=None, vcs_url=None, source_name=None,
        archive_version=None, verify_command: Optional[str] = None):
    if exclude is None:
        exclude = set()
    logging.info("Processing %s", package)
    # Can't use open_packaging_branch here, since we want to use pkg_source
    # later on.
    if "/" not in package:
        try:
            with apt_repo:
                pkg_source = apt_get_source_package(apt_repo, package)
        except NoSuchPackage:
            logging.info("%s: package not found in apt", package)
            raise PackageProcessingFailure('not-in-apt')
        if vcs_type is None or vcs_url is None:
            try:
                vcs_type, vcs_url = source_package_vcs(pkg_source)
            except KeyError:
                logging.info(
                    "%s: no declared vcs location, skipping",
                    pkg_source["Package"]
                )
                raise PackageProcessingFailure('not-in-vcs')
        if source_name is None:
            source_name = pkg_source["Package"]
        if source_name in exclude:
            raise PackageIgnored('excluded')
        if archive_version is None:
            archive_version = pkg_source["Version"]
        has_testsuite = "Testsuite" in pkg_source
    else:
        if vcs_url is None:
            vcs_url = package
        has_testsuite = None
    (location, branch_name, subpath) = split_vcs_url(vcs_url)
    if subpath is None:
        subpath = ""
    probers = select_probers(vcs_type)
    try:
        main_branch = open_branch(
            location, probers=probers, name=branch_name)
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        if debug:
            logging.exception("%s: %s", vcs_url, e)
        else:
            logging.info(
                "%s: branch unavailable: %s", vcs_url, e)
        raise PackageProcessingFailure('vcs-inaccessible')
    with Workspace(main_branch) as ws:
        if source_name is None:
            with ControlEditor(
                ws.local_tree.abspath(
                    os.path.join(subpath, "debian/control"))
            ) as ce:
                source_name = ce.source["Source"]
                try:
                    with apt_repo:
                        pkg_source = apt_get_source_package(
                            apt_repo, source_name)
                except NoSuchPackage:
                    logging.info("%s: package not found in apt", package)
                    raise PackageProcessingFailure('not-in-apt')
                archive_version = pkg_source['Version']
            has_testsuite = "Testsuite" in ce.source
        if source_name in exclude:
            raise PackageIgnored('excluded')
        if (
            autopkgtest_only
            and not has_testsuite
            and not ws.local_tree.has_filename(
                os.path.join(subpath, "debian/tests/control")
            )
        ):
            logging.info(
                "%s: Skipping, package has no autopkgtest.", source_name)
            raise PackageIgnored('no-autopkgtest')
        branch_config = ws.local_tree.branch.get_config_stack()
        if gpg_verification:
            gpg_strategy = gpg.GPGStrategy(branch_config)
            if not acceptable_keys:
                acceptable_keys = list(
                    get_maintainer_keys(gpg_strategy.context))
            gpg_strategy.set_acceptable_keys(",".join(acceptable_keys))
        else:
            gpg_strategy = None

        try:
            target_changes, tag_name = prepare_upload_package(
                ws.local_tree,
                subpath,
                source_name,
                archive_version,
                builder=builder,
                gpg_strategy=gpg_strategy,
                min_commit_age=min_commit_age,
                allowed_committers=allowed_committers,
                apt=apt_repo,
            )
        except GbpDchFailed as e:
            logging.warn("%s: 'gbp dch' failed to run: %s", source_name, e)
            raise PackageProcessingFailure('gbp-dch-failed')
        except MissingUpstreamTarball as e:
            logging.warning(
                "%s: missing upstream tarball: %s", source_name, e)
            raise PackageProcessingFailure('missing-upstream-tarball')
        except BranchRateLimited as e:
            logging.warning(
                '%s: rate limited by server (retrying after %s)',
                source_name, e.retry_after)
            raise PackageProcessingFailure('rate-limited')
        except CommitterNotAllowed as e:
            logging.warn(
                "%s: committer %s not in allowed list: %r",
                source_name,
                e.committer,
                e.allowed_committers,
            )
            raise PackageIgnored('committer-not-allowed')
        except BuildFailedError as e:
            logging.warn("%s: package failed to build: %s", source_name, e)
            raise PackageProcessingFailure('build-failed')
        except LastReleaseRevisionNotFound as e:
            logging.warn(
                "%s: Unable to find revision matching last release "
                "%s, skipping.",
                source_name,
                e.version,
            )
            raise PackageProcessingFailure('last-release-missing')
        except LastUploadMoreRecent as e:
            logging.warn(
                "%s: Last upload (%s) was more recent than VCS (%s)",
                source_name,
                e.archive_version,
                e.vcs_version,
            )
            raise PackageProcessingFailure('last-upload-not-in-vcs')
        except ChangelogParseError as e:
            logging.info("%s: Error parsing changelog: %s", source_name, e)
            raise PackageProcessingFailure('changelog-parse-error')
        except MissingChangelogError:
            logging.info("%s: No changelog found, skipping.", source_name)
            raise PackageProcessingFailure('missing-changelog')
        except GeneratedChangelogFile:
            logging.info(
                "%s: Changelog is generated and unable to update, skipping.",
                source_name)
            raise PackageProcessingFailure('generated-changelog')
        except RecentCommits as e:
            logging.info(
                "%s: Recent commits (%d days), skipping.",
                source_name, e.commit_age
            )
            raise PackageIgnored('recent-commits')
        except NoUnuploadedChanges as e:
            logging.info("%s: No unuploaded changes (%s), skipping.",
                         source_name, e.archive_version)
            raise PackageIgnored('no-unuploaded-changes')
        except NoUnreleasedChanges:
            logging.info(
                "%s: No unreleased changes, skipping.", source_name)
            raise PackageIgnored('no-unreleased-changes')
        except MissingNestedTree:
            logging.exception('missing nested tree')
            raise PackageIgnored('unsuported-nested-tree')

        if verify_command:
            try:
                subprocess.check_call([verify_command, target_changes])
            except subprocess.CalledProcessError as e:
                if e.returncode == 1:
                    raise PackageIgnored(
                        'verify-command-declined',
                        "{}: Verify command {!r} declined upload".format(
                            source_name, verify_command))
                else:
                    raise PackageProcessingFailure(
                        'verify-command-error',
                        f"{source_name}: "
                        f"Error running verify command {verify_command}: "
                        f"returncode {e.returncode}")

        tags = []
        if tag_name is not None:
            logging.info("Pushing tag %s", tag_name)
            tags.append(tag_name)
        try:
            ws.push(tags=tags)
        except PermissionDenied:
            logging.info(
                "%s: Permission denied pushing to branch, skipping.",
                source_name
            )
            raise PackageProcessingFailure('vcs-permission-denied')
        dput_changes(target_changes)
        if diff:
            sys.stdout.flush()
            ws.show_diff(sys.stdout.buffer)
            sys.stdout.buffer.flush()


def vcswatch_prescan_package(
        package, vw, exclude=None, min_commit_age=None,
        allowed_committers=None):
    vcs_url = vw["url"]
    vcs_type = vw["vcs"]
    source_name = vw["package"]
    if source_name in exclude:
        raise PackageIgnored('excluded')
    if vcs_url is None or vcs_type is None:
        raise PackageProcessingFailure('not-in-vcs')
    # TODO(jelmer): check autopkgtest_only ?
    # from debian.deb822 import Deb822
    # pkg_source = Deb822(vw["controlfile"])
    # has_testsuite = "Testsuite" in pkg_source
    if vw["commits"] == 0:
        raise PackageIgnored('no-unuploaded-changes')
    if vw['status'] == 'ERROR':
        logging.warning(
            'vcswatch: unable to access %s: %s', vw['package'], vw['error'])
        raise PackageProcessingFailure('vcs-inaccessible')
    logging.debug("vcswatch last scanned at: %s", vw["last_scan"])
    if vcs_type == 'Git' and vw["vcslog"] is not None:
        try:
            return check_git_commits(
                vw["vcslog"], min_commit_age=min_commit_age,
                allowed_committers=allowed_committers)
        except CommitterNotAllowed as e:
            logging.warn(
                "%s: committer %s not in allowed list: %r",
                source_name,
                e.committer,
                e.allowed_committers,
            )
            raise PackageIgnored('committer-not-allowed')
        except RecentCommits as e:
            logging.info(
                "%s: Recent commits (%d days), skipping.",
                source_name, e.commit_age
            )
            raise PackageIgnored('recent-commits')
    return None


def vcswatch_prescan_packages(
        packages, inc_stats: Callable[[str], None],
        exclude=None, min_commit_age=None,
        allowed_committers=None):
    logging.info('Using vcswatch to prescan %d packages', len(packages))
    import gzip
    import json
    from urllib.request import Request, urlopen

    from .. import version_string
    url = "https://qa.debian.org/data/vcswatch/vcswatch.json.gz"
    request = Request(url, headers={
        'User-Agent': "silver-platter/%s" % version_string})
    vcswatch = {
        p['package']: p
        for p in json.load(gzip.GzipFile(fileobj=urlopen(request)))}
    by_ts = {}
    failures = 0
    for package in packages:
        try:
            vw = vcswatch[package]
        except KeyError:
            continue
        try:
            ts = vcswatch_prescan_package(
                    package, vw,
                    exclude=exclude, min_commit_age=min_commit_age,
                    allowed_committers=allowed_committers)
        except PackageProcessingFailure as e:
            inc_stats(e.reason)
            failures += 1
        except PackageIgnored as e:
            inc_stats(e.reason)
        else:
            by_ts[package] = ts
    return ([k for (k, v) in
             sorted(by_ts.items(), key=lambda k: k[1] or 0, reverse=True)],
            failures,
            vcswatch)


def open_last_attempt_db():
    try:
        import tdb
        from xdg.BaseDirectory import xdg_data_home
    except ModuleNotFoundError:
        return None
    else:
        last_attempt_path = os.path.join(
            xdg_data_home, 'silver-platter', 'last-upload-attempt.tdb')
        os.makedirs(os.path.dirname(last_attempt_path), exist_ok=True)
        return tdb.open(
            last_attempt_path, tdb_flags=tdb.DEFAULT,
            flags=os.O_RDWR | os.O_CREAT, mode=0o600)


def main(argv):  # noqa: C901
    import argparse

    parser = argparse.ArgumentParser(prog="upload-pending-commits")
    parser.add_argument("packages", nargs="*")
    parser.add_argument(
        "--acceptable-keys",
        help="List of acceptable GPG keys",
        action="append",
        default=[],
        type=str,
    )
    parser.add_argument(
        "--gpg-verification",
        help="Verify GPG signatures on commits",
        action="store_true",
    )
    parser.add_argument(
        "--min-commit-age",
        help="Minimum age of the last commit, in days",
        type=int,
        default=0,
    )
    parser.add_argument("--diff", action="store_true", help="Show diff.")
    parser.add_argument(
        "--builder",
        type=str,
        help="Build command",
        default=(
            DEFAULT_BUILDER + " --source --source-only-changes "
            "--debbuildopt=-v${LAST_VERSION}"
        ),
    )
    parser.add_argument(
        "--maintainer",
        type=str,
        action="append",
        help="Select all packages maintained by specified maintainer.",
    )
    parser.add_argument(
        "--vcswatch",
        action="store_true",
        default=False,
        help="Use vcswatch to determine what packages need uploading."
    )
    parser.add_argument(
        "--exclude", type=str, action="append", default=[],
        help="Ignore source package"
    )
    parser.add_argument(
        "--autopkgtest-only",
        action="store_true",
        help="Only process packages with autopkgtests.",
    )
    parser.add_argument(
        "--allowed-committer",
        type=str,
        action="append",
        help="Require that all new commits are from specified committers",
    )
    parser.add_argument(
        "--debug",
        action="store_true")
    parser.add_argument(
        "--shuffle",
        action="store_true",
        help="Randomize order packages are processed in.")
    parser.add_argument(
        "--verify-command", type=str, default=None,
        help=("Command to verify whether upload is necessary. "
              "Should return 1 to decline, 0 to upload."))
    parser.add_argument(
        '--apt-repository', type=str,
        help='APT repository to use. Defaults to locally configured.',
        default=(
            os.environ.get('APT_REPOSITORY')
            or os.environ.get('REPOSITORIES')))
    parser.add_argument(
        '--apt-repository-key', type=str,
        help=('APT repository key to use for validation, '
              'if --apt-repository is set.'),
        default=os.environ.get('APT_REPOSITORY_KEY'))

    args = parser.parse_args(argv)

    ret = 0

    if not args.packages and not args.maintainer:
        (name, email) = get_maintainer()
        if email:
            logging.info("Processing packages maintained by %s", email)
            args.maintainer = [email]
        else:
            parser.print_usage()
            sys.exit(1)

    if not args.vcswatch:
        logging.info(
            "Use --vcswatch to only process packages for which "
            "vcswatch found pending commits."
        )

    if args.apt_repository:
        apt_repo = RemoteApt.from_string(
            args.apt_repository, args.apt_repository_key)
    else:
        apt_repo = LocalApt()

    if args.maintainer:
        if args.packages:
            parser.print_error(
                '--maintainer is incompatible with specifying package names')
        packages = select_apt_packages(
            apt_repo, args.packages, args.maintainer)
    else:
        packages = args.packages

    if not packages:
        logging.info("No packages found.")
        parser.print_usage()
        sys.exit(1)

    if args.shuffle:
        import random
        random.shuffle(packages)

    stats = {}

    def inc_stats(result):
        stats.setdefault(result, 0)
        stats[result] += 1

    if args.vcswatch:
        packages, failures, extra_data = vcswatch_prescan_packages(
            packages, inc_stats, exclude=args.exclude,
            min_commit_age=args.min_commit_age,
            allowed_committers=args.allowed_committer)
        if failures > 0:
            ret = 1
    else:
        extra_data = {}

    if len(packages) > 1:
        logging.info(
            "Uploading %d packages: %s", len(packages), ", ".join(packages))

    last_attempt = open_last_attempt_db()

    if last_attempt:
        orig_packages = list(packages)

        def last_attempt_key(p):
            try:
                t = int(last_attempt[p.encode('utf-8')])
            except KeyError:
                t = 0
            return (t, orig_packages.index(p))
        packages.sort(key=last_attempt_key)

    for package in packages:
        vcs_type = extra_data.get(package, {}).get('vcs_type')
        vcs_url = extra_data.get(package, {}).get('vcs_url')
        archive_version = extra_data.get(package, {}).get('archive_version')
        source_name = extra_data.get(package, {}).get('package')

        try:
            process_package(
                apt_repo, package,
                builder=args.builder, exclude=args.exclude,
                autopkgtest_only=args.autopkgtest_only,
                gpg_verification=args.gpg_verification,
                acceptable_keys=args.acceptable_keys,
                debug=args.debug,
                diff=args.diff, min_commit_age=args.min_commit_age,
                allowed_committers=args.allowed_committer,
                vcs_type=vcs_type, vcs_url=vcs_url,
                archive_version=archive_version,
                source_name=source_name,
                verify_command=args.verify_command)
        except PackageProcessingFailure as e:
            inc_stats(e.reason)
            ret = 1
        except PackageIgnored as e:
            inc_stats(e.reason)

        if last_attempt:
            last_attempt[package.encode('utf-8')] = b'%d' % (time.time())

    if len(packages) > 1:
        logging.info('Results:')
        for error, c in stats.items():
            logging.info('  %s: %d', error, c)

    return ret


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python
# Copyright (C) 2018-2019 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import logging
import sys
from typing import Callable, Dict, List, Optional

import silver_platter  # noqa: F401

from . import apply, batch, run, version_string


def forges_main(argv: List[str]) -> Optional[int]:
    from .proposal import forges
    parser = argparse.ArgumentParser(prog="svp forges")
    parser.parse_args(argv)

    for name, forge_cls in forges.items():
        for instance in forge_cls.iter_instances():
            print(f"{instance.base_url} ({name})")

    return None


def login_main(argv: List[str]) -> Optional[int]:
    parser = argparse.ArgumentParser(prog="svp login")
    parser.add_argument("url", help="URL of branch to work on.", type=str)
    args = parser.parse_args(argv)

    try:
        from launchpadlib import uris as lp_uris
    except ModuleNotFoundError:
        logging.warning(
            'launchpadlib is not installed, unable to log in to launchpad')
        lp_uris = []

    forge = None
    # TODO(jelmer): Don't special case various forges here
    if args.url.startswith("https://github.com"):
        forge = "github"
    for key, root in lp_uris.web_roots.items():
        if args.url.startswith(root) or args.url == root.rstrip("/"):
            forge = "launchpad"
            lp_service_root = lp_uris.service_roots[key]
    if forge is None:
        forge = "gitlab"

    if forge == "gitlab":
        from breezy.plugins.gitlab.cmds import cmd_gitlab_login

        cmd_gl = cmd_gitlab_login()
        cmd_gl._setup_outf()
        return cmd_gl.run(args.url)
    elif forge == "github":
        from breezy.plugins.github.cmds import cmd_github_login

        cmd_gh = cmd_github_login()
        cmd_gh._setup_outf()
        return cmd_gh.run()
    elif forge == "launchpad":
        from breezy.plugins.launchpad.cmds import cmd_launchpad_login

        cmd_lp = cmd_launchpad_login()
        cmd_lp._setup_outf()
        cmd_lp.run()
        from breezy.plugins.launchpad import lp_api

        lp_api.connect_launchpad(lp_service_root, version="devel")
        return None
    else:
        logging.fatal("Unknown forge %r.", forge)
        return 1


def proposals_main(argv: List[str]) -> None:
    from .proposal import iter_all_mps

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--status",
        default="open",
        choices=["open", "merged", "closed"],
        type=str,
        help="Only display proposals with this status.",
    )
    args = parser.parse_args(argv)

    for _forge, proposal, _status in iter_all_mps([args.status]):
        print(proposal.url)


subcommands: Dict[str, Callable[[List[str]], Optional[int]]] = {
    "forges": forges_main,
    "login": login_main,
    "proposals": proposals_main,
    "run": run.main,
    "apply": apply.main,
    "batch": batch.main,
}


def main(argv: Optional[List[str]] = None) -> Optional[int]:
    import breezy

    breezy.initialize()  # type: ignore
    parser = argparse.ArgumentParser(prog="svp", add_help=False)
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + version_string
    )
    parser.add_argument(
        "--help", action="store_true", help="show this help message and exit"
    )
    parser.add_argument(
        "subcommand", type=str, choices=list(subcommands.keys()))
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    args, rest = parser.parse_known_args(argv)
    if args.help:
        if args.subcommand is None:
            parser.print_help()
            parser.exit()
        else:
            rest.append("--help")
    if args.subcommand is None:
        parser.print_usage()
        return 1
    return subcommands[args.subcommand](rest)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import logging
import sys
from typing import Callable, Dict, List, Optional

import silver_platter  # noqa: F401

from . import apply as debian_apply
from . import batch as debian_batch
from . import run as debian_run
from . import uploader as debian_uploader


def main(argv: Optional[List[str]] = None) -> Optional[int]:
    import breezy

    breezy.initialize()  # type: ignore

    from ..__main__ import subcommands as main_subcommands

    subcommands: Dict[str, Callable[[List[str]], Optional[int]]] = {
        "upload-pending": debian_uploader.main,
        "apply": debian_apply.main,
        "run": debian_run.main,
        "batch": debian_batch.main,
    }

    parser = argparse.ArgumentParser(prog="debian-svp", add_help=False)
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + silver_platter.version_string,
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Be more verbose")
    parser.add_argument(
        "--help", action="store_true", help="show this help message and exit"
    )

    for name, cmd in main_subcommands.items():
        if name not in subcommands:
            subcommands[name] = cmd

    parser.add_argument(
        "subcommand", type=str, choices=list(subcommands.keys())
    )
    args, rest = parser.parse_known_args()
    if args.help:
        if args.subcommand is None:
            parser.print_help()
            parser.exit()
        else:
            rest.append("--help")
    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level, format="%(message)s")
    if args.subcommand is None:
        parser.print_usage()
        return 1
    if args.subcommand in subcommands:
        return subcommands[args.subcommand](rest)
    parser.print_usage()
    return 1


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python
# Copyright (C) 2018-2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import logging
import os
import shutil
import tempfile
from typing import Any, BinaryIO, Callable, Dict, List, Optional, Tuple, Union

from breezy.branch import Branch
from breezy.controldir import ControlDir, NoColocatedBranchSupport
from breezy.diff import show_diff_trees
from breezy.errors import DivergedBranches, NotBranchError
from breezy.revision import NULL_REVISION, RevisionID
from breezy.transport.local import LocalTransport
from breezy.tree import Tree
from breezy.workingtree import WorkingTree

from .proposal import Forge, MergeProposal, UnsupportedForge, get_forge
from .publish import (
    PublishResult,
    propose_changes,
    push_changes,
    push_derived_changes,
    publish_changes as _publish_changes,
)
from .utils import create_temp_sprout, full_branch_url

__all__ = [
    "Workspace",
]


logger = logging.getLogger(__name__)


def fetch_colocated(controldir: ControlDir, from_controldir: ControlDir,
                    additional_colocated_branches: Dict[str, str]):
    logger.debug(
        "Fetching colocated branches: %r",
        additional_colocated_branches,
    )
    for (from_branch_name,
         to_branch_name) in additional_colocated_branches.items():
        try:
            remote_colo_branch = from_controldir.open_branch(
                name=from_branch_name
            )
        except (NotBranchError, NoColocatedBranchSupport):
            continue
        controldir.push_branch(
            name=to_branch_name, source=remote_colo_branch, overwrite=True
        )


class Workspace:
    """Workspace for creating changes to a branch.

    Args:
        main_branch: The upstream branch
        resume_branch: Optional in-progress branch that we previously made
            changes on, and should ideally continue from.
        resume_branch_additional_colocated_branches:
            Additional list of colocated branches to fetch
        cached_branch: Branch to copy revisions from, if possible.
        local_tree: The tree the user can work in
    """

    _destroy: Optional[Callable[[], None]]
    local_tree: WorkingTree
    main_branch_revid: Optional[RevisionID]
    main_colo_revid: Dict[Optional[str], RevisionID]
    additional_colocated_branches: Dict[str, str]
    resume_branch_additional_colocated_branches: Optional[Dict[str, str]]

    @classmethod
    def from_url(cls, url, **kwargs):
        return cls(main_branch=Branch.open(url), **kwargs)

    def __init__(
        self,
        main_branch: Optional[Branch],
        *,
        resume_branch: Optional[Branch] = None,
        cached_branch: Optional[Branch] = None,
        additional_colocated_branches:
            Optional[Union[List[str], Dict[str, str]]] = None,
        resume_branch_additional_colocated_branches:
            Optional[Union[List[str], Dict[str, str]]] = None,
        dir: Optional[str] = None,
        path: Optional[str] = None,
        format=None
    ) -> None:
        self.main_branch = main_branch
        self.main_branch_revid = None
        self.refreshed = False
        self.cached_branch = cached_branch
        self.resume_branch = resume_branch
        if additional_colocated_branches is None:
            additional_colocated_branches = {}
        elif isinstance(additional_colocated_branches, list):
            additional_colocated_branches = {
                k: k for k in additional_colocated_branches}
        self.additional_colocated_branches = additional_colocated_branches
        if isinstance(resume_branch_additional_colocated_branches, list):
            resume_branch_additional_colocated_branches = {
                k: k for k in resume_branch_additional_colocated_branches}
        self.resume_branch_additional_colocated_branches = (
            resume_branch_additional_colocated_branches)
        self._destroy = None
        self._dir = dir
        self._path = path
        self._format = format

    @property
    def path(self):
        return self.local_tree.abspath('.')

    def __str__(self) -> str:
        if self._path is None:
            if self.main_branch is None:
                return "Workspace"
            else:
                return "Workspace for %s" % full_branch_url(self.main_branch)
        else:
            if self.main_branch:
                return "Workspace for {} at {}".format(
                    full_branch_url(self.main_branch),
                    self._path)
            else:
                return "Workspace at %s" % self._path

    def __repr__(self) -> str:
        return (
            "{}({!r}, resume_branch={!r}, cached_branch={!r}, "
            "additional_colocated_branches={!r}, "
            "resume_branch_additional_colocated_branches={!r}, "
            "dir={!r}, path={!r})".format(
                type(self).__name__,
                self.main_branch,
                self.resume_branch,
                self.cached_branch,
                self.additional_colocated_branches,
                self.resume_branch_additional_colocated_branches,
                self._dir,
                self._path,
            )
        )

    def _inverse_additional_colocated_branches(self):
        return [(to_name, from_name)
                for from_name, to_name in
                self.additional_colocated_branches.items()]

    def __enter__(self) -> Any:
        sprout_base = None
        for (sprout_base, sprout_coloc) in [  # noqa: B007
                (self.cached_branch, self.additional_colocated_branches),
                (self.resume_branch,
                    self.resume_branch_additional_colocated_branches),
                (self.main_branch, self.additional_colocated_branches)]:
            if sprout_base:
                break

        if sprout_base is None:
            logger.debug(
                'Creating new empty tree with format %r', self._format)
            if self._path is not None:
                os.mkdir(self._path)
                td = self._path
            else:
                td = tempfile.mkdtemp(dir=self._dir)
                self._destroy = lambda: shutil.rmtree(td)
            self.local_tree = ControlDir.create_standalone_workingtree(
                td, format=self._format)
        else:
            logger.debug("Creating sprout from %r", sprout_base)
            self.local_tree, self._destroy = create_temp_sprout(
                sprout_base,
                sprout_coloc,
                dir=self._dir,
                path=self._path,
            )
        if self.main_branch:
            self.main_branch_revid = self.main_branch.last_revision()
        else:
            self.main_branch_revid = NULL_REVISION
        self.main_colo_revid = {}
        self.refreshed = False
        if self.main_branch:
            for from_name, to_name in (
                    self.additional_colocated_branches.items()):
                try:
                    branch = self.main_branch.controldir.open_branch(
                        name=from_name)  # type: ignore
                except (NotBranchError, NoColocatedBranchSupport):
                    continue
                self.main_colo_revid[to_name] = branch.last_revision()

            if self.cached_branch:
                logger.debug(
                    "Pulling in missing revisions from resume/main branch %r",
                    self.resume_branch or self.main_branch,
                )
                self.local_tree.pull(
                    self.resume_branch or self.main_branch, overwrite=True
                )
            # At this point, we're either on the tip of the main branch or the
            # tip of the resume branch
            if self.resume_branch:
                # If there's a resume branch at play, make sure it's derived
                # from the main branch *or* reset back to the main branch.
                logger.debug(
                    "Pulling in missing revisions from main branch %r",
                    self.main_branch
                )
                try:
                    self.local_tree.pull(self.main_branch, overwrite=False)
                except DivergedBranches:
                    logger.info("restarting branch")
                    self.refreshed = True
                    self.resume_branch = None
                    self.resume_branch_additional_colocated_branches = None
                    self.local_tree.pull(self.main_branch, overwrite=True)
                    fetch_colocated(
                        self.local_tree.branch.controldir,
                        self.main_branch.controldir,
                        self.additional_colocated_branches)
                else:
                    fetch_colocated(
                        self.local_tree.branch.controldir,
                        self.main_branch.controldir,
                        self.additional_colocated_branches)
                    if self.resume_branch_additional_colocated_branches:
                        fetch_colocated(
                            self.local_tree.branch.controldir,
                            self.resume_branch.controldir,
                            self.resume_branch_additional_colocated_branches)
                        self.additional_colocated_branches.update(
                            self.resume_branch_additional_colocated_branches)
            else:
                fetch_colocated(
                    self.local_tree.branch.controldir,
                    self.main_branch.controldir,
                    self.additional_colocated_branches)

        self.base_revid = self.local_tree.last_revision()
        return self

    def defer_destroy(self) -> Optional[Callable[[], None]]:
        ret = self._destroy
        self._destroy = None
        return ret

    def changes_since_main(self) -> bool:
        return self.local_tree.branch.last_revision() != self.main_branch_revid

    def changes_since_base(self) -> bool:
        return self.base_revid != self.local_tree.branch.last_revision()

    def any_branch_changes(self):
        """Have any branch changes at all been made?

        Includes changes that already existed in the resume branch.
        """
        return any(br != r for name, br, r in self.result_branches())

    def result_branches(self) -> List[
            Tuple[Optional[str], Optional[RevisionID], Optional[RevisionID]]]:
        """Return a list of branches that has changed.

        Returns:
           List of tuples with (branch name, old revid, new revid)
        """
        branches = [
            (self.main_branch.name if self.main_branch else '',
             self.main_branch_revid,  # type: ignore
             self.local_tree.last_revision())]
        for from_name, to_name in self.additional_colocated_branches.items():
            to_revision: Optional[RevisionID]
            try:
                to_branch = self.local_tree.controldir.open_branch(
                    name=to_name)
            except NoColocatedBranchSupport:
                continue
            except NotBranchError:
                to_revision = None
            else:
                to_revision = to_branch.last_revision()
            from_revision = self.main_colo_revid.get(from_name)
            if from_revision is None and to_revision is None:
                continue
            branches.append((from_name, from_revision, to_revision))
        names = [name for (name, from_rev, to_rev) in branches]
        assert len(names) == len(set(names)), \
            "Duplicate result branches: %r" % branches
        return branches

    def push_tags(
            self,
            tags: Dict[str, RevisionID],
            *,
            forge: Optional[Forge] = None):
        if not self.main_branch:
            raise RuntimeError('no main branch known')
        return self.push(
            forge=forge,
            tags=tags,
            stop_revision=self.main_branch.last_revision())

    def push(
        self,
        *,
        forge: Optional[Forge] = None,
        tags: Optional[Dict[str, RevisionID]] = None,
        stop_revision: Optional[RevisionID] = None,
    ) -> None:
        if not self.main_branch:
            raise RuntimeError('no main branch known')
        if forge is None:
            try:
                forge = get_forge(self.main_branch)
            except UnsupportedForge:
                if not isinstance(
                        self.main_branch.control_transport, LocalTransport):
                    logging.warning(
                        'Unable to find forge for %s to determine push url, '
                        'trying anyway.', self.main_branch.user_url)
                forge = None
        return push_changes(
            self.local_tree.branch,
            self.main_branch,
            forge=forge,
            additional_colocated_branches=(
                self._inverse_additional_colocated_branches()),
            tags=tags,
            stop_revision=stop_revision,
        )

    def propose(
        self,
        name: str,
        description: str,
        *,
        target_branch: Optional[Branch] = None,
        forge: Optional[Forge] = None,
        existing_proposal: Optional[MergeProposal] = None,
        overwrite_existing: Optional[bool] = None,
        labels: Optional[List[str]] = None,
        commit_message: Optional[str] = None,
        title: Optional[str] = None,
        reviewers: Optional[List[str]] = None,
        tags: Optional[Union[Dict[str, RevisionID], List[str]]] = None,
        owner: Optional[str] = None,
        allow_collaboration: bool = False,
        stop_revision: Optional[RevisionID] = None,
    ) -> Tuple[MergeProposal, bool]:
        if target_branch is None:
            target_branch = self.main_branch
        if not target_branch:
            raise RuntimeError('no main branch known')
        if forge is None:
            forge = get_forge(target_branch)
        return propose_changes(
            self.local_tree.branch,
            target_branch,
            forge=forge,
            name=name,
            mp_description=description,
            resume_branch=self.resume_branch,
            resume_proposal=existing_proposal,
            overwrite_existing=(overwrite_existing or False),
            labels=labels,
            commit_message=commit_message,
            title=title,
            reviewers=reviewers,
            owner=owner,
            additional_colocated_branches=(
                self._inverse_additional_colocated_branches()),
            tags=tags,
            allow_collaboration=allow_collaboration,
            stop_revision=stop_revision,
        )

    def push_derived(
        self,
        name: str,
        *,
        target_branch: Optional[Branch] = None,
        forge: Optional[Forge] = None,
        overwrite_existing: Optional[bool] = False,
        owner: Optional[str] = None,
        tags: Optional[Union[Dict[str, RevisionID], List[str]]] = None,
        stop_revision: Optional[RevisionID] = None,
    ) -> Tuple[Branch, str]:
        """Push a derived branch.

        Args:
          name: Branch name
          forge: Optional forge to use
          overwrite_existing: Whether to overwrite an existing branch
          tags: Tags list to push
          owner: Owner name
        Returns:
          tuple with remote_branch and public_branch_url
        """
        if target_branch is None:
            target_branch = self.main_branch
        if not self.main_branch:
            raise RuntimeError('no main branch known')
        if forge is None:
            forge = get_forge(self.main_branch)
        return push_derived_changes(
            self.local_tree.branch,
            self.main_branch,
            forge,
            name,
            overwrite_existing=overwrite_existing,
            owner=owner,
            tags=tags,
            stop_revision=stop_revision,
        )

    def publish_changes(self, *,
                        target_branch: Optional[Branch] = None,
                        **kwargs) -> PublishResult:
        """Publish a set of changes."""
        if target_branch is None:
            target_branch = self.main_branch
        if not target_branch:
            raise RuntimeError('no main branch known')
        return _publish_changes(
            self.local_tree.branch,
            target_branch,
            self.resume_branch,
            **kwargs
        )

    def base_tree(self) -> Tree:
        return self.local_tree.branch.repository.revision_tree(self.base_revid)

    def show_diff(
        self, outf: BinaryIO, old_label: str = "old/", new_label: str = "new/"
    ) -> None:
        base_tree = self.base_tree()
        show_diff_trees(
            base_tree,
            self.local_tree.basis_tree(),
            outf,
            old_label=old_label,
            new_label=new_label,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._destroy:
            self._destroy()
            self._destroy = None
        return False



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for updating with a script."""

import argparse
import logging
import os
import sys
from typing import List, Optional

from breezy import osutils
from breezy.urlutils import InvalidURL

import silver_platter  # noqa: F401

from ..candidates import Candidate, CandidateList
from ..proposal import (
    ForgeLoginRequired,
    MergeProposal,
    UnsupportedForge,
    enable_tag_pushing,
    find_existing_proposed,
    get_forge,
)
from ..publish import SUPPORTED_MODES, InsufficientChangesForNewProposal
from ..utils import (
    BranchMissing,
    BranchUnavailable,
    BranchUnsupported,
    full_branch_url,
    open_branch,
)
from . import (
    DEFAULT_BUILDER,
    BuildFailedError,
    MissingUpstreamTarball,
    Workspace,
    build,
)
from .apply import (
    MissingChangelog,
    ScriptFailed,
    ScriptMadeNoChanges,
    ScriptNotFound,
    install_built_package,
    script_runner,
)


def derived_branch_name(script: str) -> str:
    return os.path.splitext(osutils.basename(script.split(" ")[0]))[0]


def apply_and_publish(  # noqa: C901
        url: str, name: str, command: str, mode: str,
        subpath: str = '',
        commit_pending: Optional[bool] = None,
        labels: Optional[List[str]] = None, diff: bool = False,
        verify_command: Optional[str] = None,
        derived_owner: Optional[str] = None,
        refresh: bool = False, allow_create_proposal=None,
        get_commit_message=None, get_title=None, get_description=None,
        build_verify=False, builder=DEFAULT_BUILDER, install=False,
        build_target_dir=None, update_changelog: Optional[bool] = None,
        preserve_repositories: bool = False):
    try:
        main_branch = open_branch(url)
    except (BranchUnavailable, BranchMissing, BranchUnsupported) as e:
        logging.fatal("%s: %s", url, e)
        return 1
    except InvalidURL as e:
        logging.fatal('%s: %s', url, e)
        return 1

    overwrite = False

    try:
        forge = get_forge(main_branch)
    except UnsupportedForge as e:
        if mode != "push":
            raise
        # We can't figure out what branch to resume from when there's no forge
        # that can tell us.
        resume_branch = None
        existing_proposals: Optional[List[MergeProposal]] = []
        logging.warn(
            "Unsupported forge (%s), will attempt to push to %s",
            e,
            full_branch_url(main_branch),
        )
    except ForgeLoginRequired as e:
        logging.error(
            '%s: Forge login required: %s', full_branch_url(main_branch), e)
        return 1
    else:
        (resume_branch, resume_overwrite,
         existing_proposals) = find_existing_proposed(
             main_branch, forge, name, owner=derived_owner)
        if resume_overwrite is not None:
            overwrite = resume_overwrite
    if refresh:
        resume_branch = None

    if existing_proposals and len(existing_proposals) > 1:
        logging.warning(
            'Multiple open merge proposals for branch at %s: %r',
            resume_branch.user_url,  # type: ignore
            [mp.url for mp in existing_proposals])
        existing_proposal = existing_proposals[0]
        logging.info('Updating just %s', existing_proposal.url)
    else:
        existing_proposal = None

    with Workspace(main_branch, resume_branch=resume_branch) as ws:
        try:
            result = script_runner(
                ws.local_tree, command, commit_pending,
                update_changelog=update_changelog)
        except MissingChangelog as e:
            logging.error("No debian changelog (%s) present", e.args[0])
            return 1
        except ScriptMadeNoChanges:
            logging.error("Script did not make any changes.")
            return 1
        except ScriptFailed:
            logging.error("Script failed to run.")
            return 1
        except ScriptNotFound:
            logging.error("Script could not be found.")
            return 1

        if build_verify or install:
            try:
                build(ws.local_tree, subpath, builder=builder,
                      result_dir=build_target_dir)
            except BuildFailedError:
                logging.info("%s: build failed", result.source)
                return False
            except MissingUpstreamTarball:
                logging.info(
                    "%s: unable to find upstream source", result.source)
                return False

        if install:
            install_built_package(ws.local_tree, subpath, build_target_dir)

        enable_tag_pushing(ws.local_tree.branch)

        try:
            publish_result = ws.publish_changes(
                mode,
                name,
                get_proposal_description=(
                    lambda df, ep: get_description(result, df, ep)),
                get_proposal_commit_message=(
                    lambda ep: get_commit_message(result, ep)),
                get_proposal_title=(
                    lambda ep: get_title(result, ep)),
                allow_create_proposal=(
                    lambda: allow_create_proposal(result)),
                forge=forge,
                labels=labels,
                overwrite_existing=overwrite,
                derived_owner=derived_owner,
                existing_proposal=(
                    existing_proposals[0] if existing_proposals else None),
            )
        except UnsupportedForge as e:
            logging.error(
                "No known supported forge for %s. Run 'svp login'?",
                full_branch_url(e.branch),
            )
            return 1
        except InsufficientChangesForNewProposal:
            logging.info('Insufficient changes for a new merge proposal')
            return 0
        except ForgeLoginRequired as e:
            logging.error(
                "Credentials for hosting site at %r missing. "
                "Run 'svp login'?",
                e.forge.base_url,
            )
            return 1

        if publish_result.proposal:
            if publish_result.is_new:
                logging.info("Merge proposal created.")
            else:
                logging.info("Merge proposal updated.")
            if publish_result.proposal.url:
                logging.info("URL: %s", publish_result.proposal.url)
            logging.info(
                "Description: %s", publish_result.proposal.get_description())

        if diff:
            ws.show_diff(sys.stdout.buffer)

        if preserve_repositories:
            ws.defer_destroy()
            logging.info(
                'Workspace preserved in %s', ws.local_tree.abspath(ws.subpath))


def main(argv: List[str]) -> Optional[int]:  # noqa: C901
    parser = argparse.ArgumentParser()
    parser.add_argument("url", help="URL of branch to work on.", type=str)
    parser.add_argument(
        "--command", help="Path to script to run.", type=str)
    parser.add_argument(
        "--derived-owner", type=str, default=None,
        help="Owner for derived branches."
    )
    parser.add_argument(
        "--refresh",
        action="store_true",
        help="Refresh changes if branch already exists",
    )
    parser.add_argument(
        "--label", type=str, help="Label to attach",
        action="append", default=[]
    )
    parser.add_argument(
        "--preserve-repositories", action="store_true",
        help="Preserve temporary repositories.")

    parser.add_argument(
        "--name", type=str, help="Proposed branch name", default=None)
    parser.add_argument(
        "--diff", action="store_true", help="Show diff of generated changes."
    )
    parser.add_argument(
        "--mode",
        help="Mode for pushing",
        choices=SUPPORTED_MODES,
        default="propose",
        type=str,
    )
    parser.add_argument(
        "--commit-pending",
        help="Commit pending changes after script.",
        choices=["yes", "no", "auto"],
        default=None,
        type=str,
    )
    parser.add_argument(
        "--build-verify",
        help="Build package to verify it.",
        dest="build_verify",
        action="store_true",
    )
    parser.add_argument(
        "--builder",
        default=DEFAULT_BUILDER,
        type=str,
        help="Build command to use when verifying build.",
    )
    parser.add_argument(
        "--build-target-dir",
        type=str,
        help=(
            "Store built Debian files in specified directory "
            "(with --build-verify)"
        ),
    )
    parser.add_argument(
        "--install", "-i",
        action="store_true",
        help="Install built package (implies --build-verify)")

    parser.add_argument(
        "--recipe", type=str, help="Recipe to use.")
    parser.add_argument(
        "--candidates", type=str, help="File with candidate list.")
    parser.add_argument(
        "--no-update-changelog",
        action="store_false",
        default=None,
        dest="update_changelog",
        help="do not update the changelog",
    )
    parser.add_argument(
        "--update-changelog",
        action="store_true",
        dest="update_changelog",
        help="force updating of the changelog",
        default=None,
    )

    args = parser.parse_args(argv)

    if args.recipe:
        from ..recipe import Recipe
        recipe = Recipe.from_path(args.recipe)
    else:
        recipe = None

    candidates = []

    if args.url:
        candidates = [Candidate(url=args.url)]

    if args.candidates:
        candidatelist = CandidateList.from_path(args.candidates)
        candidates.extend(candidatelist)

    if args.commit_pending:
        commit_pending = {
            "auto": None, "yes": True, "no": False}[args.commit_pending]
    elif recipe:
        commit_pending = recipe.commit_pending
    else:
        commit_pending = None

    if args.command:
        command = args.command
    elif recipe and recipe.command:
        command = recipe.command
    else:
        parser.error('No command specified.')

    if args.name is not None:
        name = args.name
    elif recipe and recipe.name:
        name = recipe.name
    else:
        name = derived_branch_name(command)

    refresh = args.refresh

    if recipe and not recipe.resume:
        refresh = True

    def allow_create_proposal(result):
        if result.value is None:
            return True
        if recipe.propose_threshold is not None:
            return result.value >= recipe.propose_threshold
        return True

    def get_commit_message(result, existing_proposal):
        if recipe:
            return recipe.render_merge_request_commit_message(result.context)
        if existing_proposal is not None:
            return existing_proposal.get_commit_message()
        return None

    def get_title(result, existing_proposal):
        if recipe:
            return recipe.render_merge_request_title(result.context)
        if existing_proposal is not None:
            return existing_proposal.get_title()
        return None

    def get_description(result, description_format, existing_proposal):
        if recipe:
            description = recipe.render_merge_request_description(
                description_format, result.context)
            if description:
                return description
        if result.description is not None:
            return result.description
        if existing_proposal is not None:
            return existing_proposal.get_description()
        raise ValueError("No description available")

    retcode = 0

    for candidate in candidates:
        if apply_and_publish(
                candidate.url, name=name, command=command, mode=args.mode,
                subpath=candidate.subpath,
                commit_pending=commit_pending,
                labels=args.label, diff=args.diff,
                derived_owner=args.derived_owner, refresh=refresh,
                allow_create_proposal=allow_create_proposal,
                get_commit_message=get_commit_message,
                get_title=get_title,
                get_description=get_description,
                build_verify=args.build_verify, builder=args.builder,
                install=args.install, build_target_dir=args.build_target_dir,
                update_changelog=args.update_changelog,
                preserve_repositories=args.preserve_repositories):
            retcode = 1

    return retcode


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/python
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy.tests import TestCaseWithTransport

from silver_platter.run import ScriptMadeNoChanges, script_runner


class ScriptRunnerTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree("tree")

        with open("foo.sh", "w") as f:
            f.write(
                """\
#!/bin/sh
echo Foo > bar
echo "Some message"
brz add --quiet bar
"""
            )
        os.chmod("foo.sh", 0o755)

    def test_simple_with_commit(self):
        result = script_runner(
            self.tree, os.path.abspath("foo.sh"), commit_pending=True
        )
        self.assertEqual(result.description, "Some message\n")

    def test_simple_with_autocommit(self):
        result = script_runner(self.tree, os.path.abspath("foo.sh"))
        r = self.tree.branch.repository.get_revision(self.tree.last_revision())
        self.assertEqual(r.message, "Some message\n")
        self.assertEqual(result.description, "Some message\n")

    def test_simple_with_autocommit_and_script_commits(self):
        with open("foo.sh", "w") as f:
            f.write(
                """\
#!/bin/sh
echo Foo > bar
echo "Some message"
brz add --quiet bar
brz commit --quiet -m blah
"""
            )
        os.chmod("foo.sh", 0o755)
        result = script_runner(self.tree, os.path.abspath("foo.sh"))
        rev = self.tree.branch.repository.get_revision(
            self.tree.last_revision())
        self.assertEqual(rev.message, "blah")
        self.assertEqual(result.description, "Some message\n")

    def test_simple_without_commit(self):
        self.assertRaises(
            ScriptMadeNoChanges,
            script_runner,
            self.tree,
            os.path.abspath("foo.sh"),
            commit_pending=False,
        )

    def test_no_changes(self):
        with open("foo.sh", "w") as f:
            f.write(
                """\
#!/bin/sh
echo "Some message"
"""
            )
        self.assertRaises(
            ScriptMadeNoChanges,
            script_runner,
            self.tree,
            os.path.abspath("foo.sh"),
            commit_pending=True,
        )



#!/usr/bin/python3
from setuptools import setup
from setuptools_rust import RustExtension, Binding

setup(
    rust_extensions=[RustExtension(
        "silver_platter._svp_rs", "crates/svp-py/Cargo.toml",
        binding=Binding.PyO3)],)



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy.bzr import RemoteBzrProber
from breezy.git import RemoteGitProber
from breezy.tests import TestCase

from silver_platter.probers import UnsupportedVCSProber, select_probers


class SelectProbersTests(TestCase):
    def test_none(self):
        self.assertIs(None, select_probers())
        self.assertIs(None, select_probers(None))

    def test_bzr(self):
        self.assertEqual([RemoteBzrProber], select_probers("bzr"))

    def test_git(self):
        self.assertEqual([RemoteGitProber], select_probers("git"))

    def test_unsupported(self):
        self.assertEqual([UnsupportedVCSProber("foo")], select_probers("foo"))



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from contextlib import suppress
from typing import Iterator, List, Optional, Tuple

import breezy.plugins.github  # noqa: F401
import breezy.plugins.gitlab  # noqa: F401
import breezy.plugins.launchpad  # noqa: F401
from breezy.branch import Branch
from breezy.errors import PermissionDenied
from breezy.forge import (
    Forge,
    ForgeLoginRequired,
    MergeProposal,
    NoSuchProject,
    SourceNotDerivedFromTarget,
    UnsupportedForge,
    forges,
    get_forge,
    iter_forge_instances,
)
from breezy.merge_directive import MergeDirective, MergeDirective2
from breezy.transport import Transport

from .publish import (
    SUPPORTED_MODES,
    EmptyMergeProposal,
    check_proposal_diff,
    find_existing_proposed,
    propose_changes,
    push_changes,
    push_derived_changes,
)
from .utils import full_branch_url, open_branch

__all__ = [
    "ForgeLoginRequired",
    "UnsupportedForge",
    "PermissionDenied",
    "NoSuchProject",
    "get_forge",
    "forges",
    "iter_all_mps",
    "push_changes",
    "SUPPORTED_MODES",
    "push_derived_changes",
    "propose_changes",
    "check_proposal_diff",
    "EmptyMergeProposal",
    "find_existing_proposed",
]

if SourceNotDerivedFromTarget is not None:
    __all__.append("SourceNotDerivedFromTarget")


def enable_tag_pushing(branch: Branch) -> None:
    stack = branch.get_config()
    stack.set_user_option("branch.fetch_tags", True)


def merge_directive_changes(
    local_branch: Branch,
    main_branch: Branch,
    forge: Forge,
    name: str,
    message: str,
    include_patch: bool = False,
    include_bundle: bool = False,
    overwrite_existing: bool = False,
) -> MergeDirective:
    import time

    from breezy import osutils

    remote_branch, public_branch_url = forge.publish_derived(
        local_branch, main_branch, name=name, overwrite=overwrite_existing
    )
    public_branch = open_branch(public_branch_url)
    return MergeDirective2.from_objects(
        repository=local_branch.repository,
        revision_id=local_branch.last_revision(),
        time=time.time(),
        timezone=osutils.local_time_offset(),
        target_branch=main_branch,
        public_branch=public_branch,
        include_patch=include_patch,
        include_bundle=include_bundle,
        message=message,
        base_revision_id=main_branch.last_revision(),
    )


def iter_all_mps(
    statuses: Optional[List[str]] = None,
) -> Iterator[Tuple[Forge, MergeProposal, str]]:
    """Iterate over all existing merge proposals."""
    if statuses is None:
        statuses = ["open", "merged", "closed"]
    for instance in iter_forge_instances():
        for status in statuses:
            with suppress(ForgeLoginRequired):
                for mp in instance.iter_my_proposals(status=status):
                    yield instance, mp, status


def iter_conflicted(
    branch_name: str,
) -> Iterator[Tuple[str, Branch, str, Branch, Forge, MergeProposal, bool]]:
    """Find conflicted branches owned by the current user.

    Args:
      branch_name: Branch name to search for
    """
    possible_transports: List[Transport] = []
    for forge, mp, _status in iter_all_mps(["open"]):
        try:
            if mp.can_be_merged():
                continue
        except (NotImplementedError, AttributeError):
            # TODO(jelmer): Check some other way that the branch is conflicted?
            continue
        main_branch = open_branch(
            mp.get_target_branch_url(), possible_transports=possible_transports
        )
        resume_branch = open_branch(
            mp.get_source_branch_url(), possible_transports=possible_transports
        )
        if resume_branch.name != branch_name and not (  # type: ignore
            not resume_branch.name
            and resume_branch.user_url.endswith(branch_name)  # type: ignore
        ):
            continue
        # TODO(jelmer): Find out somehow whether we need to modify a subpath?
        subpath = ""
        yield (
            full_branch_url(resume_branch),
            main_branch,
            subpath,
            resume_branch,
            forge,
            mp,
            True,
        )
